function [M,X,Y,Z,N,h,B] = makepsimages(surface,L,A,r,s,D,reflectau,intens,autoexp,atten)
%MAKEPSIMAGE generate a synthetic images matrix.
%   input
%	f, fx, fy: function and partial derivatives for the observed object
%	r, s: images dimensions
%	A: horizontal size of each image
%	L: vector/matrix containing the lights position 
%	D: albedo matrix (default ones(r,s))
%	reflectau: tolerance for reflection (0 for no reflection)
%		   if it has 2 components, the first one is the tolerance and
%		   the second one an amplification factor for reflection
%	intens: light intensity (default ones(q,1))
%	autoexp: automatic or manual camera exposure (default 1=yes)
%	atten: selects the attenuation factor; see paper (default 0)
%   output
%	M: image of the object
%	X, Y, Z: object points coordinates
%	N: normal matrix
%	h,B: pixel size and vertical size of each image
%   if L is a matrix, an image for each column of L is produces and 
%   the images are vectorized in the columns of M.
%
%   See also psdatagen

%   E. Crabu & G. Rodriguez
%   Email: elisa.crabu@unica.it, rodriguez@unica.it

%   Last revised July, 2024

% Inizialization 
p = r*s;
nc = size(L,1);
q = size(L,2);

if nargin<10 || isempty(atten), atten = 0; end
if nargin<9 || isempty(autoexp), autoexp = 1; end
if nargin<8 || isempty(intens), intens = ones(q,1); end
if nargin<7 || isempty(reflectau), reflectau = 0; end
if nargin<6 || isempty(D), D = makealbedo(1,r,s); end
if nargin<5 || isempty(s), r = 101; s = 101; end
if nargin<3, error('Missing input parameters.'), end

% Set reflection parameters
if length(reflectau)>1, reflfact = reflectau(2); else, reflfact = 1; end
reflectau = reflectau(1);

% Set matrix L based on light distance
if (nc==3) || ( (nc==4) && all(L(4,:)==0) )
	lightsatinf = 1;
elseif (nc==4) && all(L(4,:)~=0)
	lightsatinf = 0;
	if any(L(4,:)~=1)	% normalize homogeneous coordinates
		L = L * diag(1./L(4,:));
	end
else
	error('error in matrix L')
end
L = L(1:3,:);

% If surface is discrete, reset r,s
if isfield(surface,'Z')
	[s,r] = size(surface.Z);
end

% Check if albedo is in color
if size(D,3)>1, albcol = 1; else albcol = 0; end

% Set the parameters
h = A/(r-1);
B = (s-1)*h;

% Compute surface and gradients
if isfield(surface,'f')		% if surface is symbolic, sample it
	x = linspace(-A/2,A/2,r);
	y = linspace(-B/2,B/2,s);
	[X,Y] = meshgrid(x,y);
	if ~lightsatinf || nargout>3, Z = surface.f(X,Y); end
	P = surface.fx(X,Y);
	Q = surface.fy(X,Y);
elseif isfield(surface,'Z')	% else if surface is sampled, differentiate it
	X = surface.X;
	Y = surface.Y;
	Z = surface.Z;
	P = zeros(s,r);
	Q = zeros(s,r);
	P(:,2:r-1) = (Z(:,3:r)-Z(:,1:r-2))/2/h;
	P(:,1) = -Z(:,1:3)*[3;-4;1]/2/h;
	P(:,r) = Z(:,r-2:r)*[1;-4;3]/2/h;
	Q(2:s-1,:) = (Z(3:s,:)-Z(1:s-2,:))/2/h;
	Q(1,:) = -[3,-4,1]*Z(1:3,:)/2/h;
	Q(s,:) = [1,-4,3]*Z(s-2:s,:)/2/h;
else
	error('Wrong surface.')
end
K = 1./sqrt(1+P.^2+Q.^2);
P = -P.*K;
Q = -Q.*K;
N = [P(:) Q(:) K(:)]';

if albcol
	M = zeros(p,q,3);
else
	M = zeros(p,q);
end
RI = false(p,q);	% used for reflections

if lightsatinf
	for i = 1:q
		T = N' * (intens(i)*L(:,i));
		if albcol
			M(:,i,1) = max(reshape(D(:,:,1),p,1).*T,0);
			M(:,i,2) = max(reshape(D(:,:,2),p,1).*T,0);
			M(:,i,3) = max(reshape(D(:,:,3),p,1).*T,0);
		else
			M(:,i) = max(D(:).*T,0);
		end
		if autoexp
			% normalize images
			% this assumes that exposure is automatic, that is, 
			% the maximum intensity in each image is 1
			M(:,i,:) = M(:,i,:)/max(max(squeeze(M(:,i,:))));
		end
		if reflectau
			%RZ = N*diag(2*N'*L(:,i)) - (L(:,i)+[0;0;1])*ones(1,p); 
			%RI(:,i) = vecnorm(RZ,inf)'<reflectau;
			R = N*spdiags(2*N'*L(:,i),0,p,p) - L(:,i)*ones(1,p); 
			RI(:,i) = acos(R(3,:)'./vecnorm(R)') < reflectau;
		end	
	end
else
	NS = sparse( [1:3*p]', kron([1:p]',[1;1;1]), N(:));
	W = [X(:) Y(:) Z(:)]';
	for i = 1:q
		Vi = L(:,i)*ones(1,p) - W;	% light seen from each point
		if atten	% see 9.7 in https://learnwebgl.brown37.net/
			% theoretical distance attenuation
			attfact = 1./vecnorm(Vi).^2;
		else
			% distance attenuation in practice
			attfact = 1./vecnorm(Vi);
		end
		V = intens(i) * (Vi * spdiags(attfact',0,p,p));
		T = NS'*V(:);
		if albcol
			M(:,i,1) = max(reshape(D(:,:,1),p,1).*T,0);
			M(:,i,2) = max(reshape(D(:,:,2),p,1).*T,0);
			M(:,i,3) = max(reshape(D(:,:,3),p,1).*T,0);
		else
			M(:,i) = max(D(:).*T,0);
		end
		if autoexp
			% normalize images
			% this assumes that exposure is automatic, that is, 
			% the maximum intensity in each image is 1
			M(:,i,:) = M(:,i,:)/max(max(squeeze(M(:,i,:))));
		end
		if reflectau
			%RZ = N*diag(2*NS'*Vi(:)) - (Vi+[0;0;1]*ones(1,p));
			%RI(:,i) = vecnorm(RZ,inf)'<reflectau;
			R = N*spdiags(2*NS'*Vi(:),0,p,p) - Vi;
			RI(:,i) = acos(R(3,:)'./vecnorm(R)') < reflectau;
		end	
	end
end

if ~autoexp
	% normalize images
	% this assumes that exposure is fixed, so that the maximum intensity 
	% across all images is one
	M = M/max(abs(M(:)));
end

if reflectau
	for i = 1:q
		M(:,i,:) = max(M(:,i,:),reflfact*RI(:,i));
	end
end

